/*
 * input/output for any type of vector, matrix, tensor < vmtio.c >
 *
 * VMT: Vector Matrix Tensor I/O
 *
 * Copyright (C) 1999 Taiji Yamada
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "vmtio.h"
#include "vmtlib.h"

static int fccode_sizes[] = {
  0,
  sizeof(char),
  sizeof(wchar_t),
  sizeof(char *),
  sizeof(wchar_t *),
  sizeof(int),
  sizeof(short int),
  sizeof(long int),
  sizeof(unsigned int),
  sizeof(unsigned short int),
  sizeof(unsigned long int),
  sizeof(float),
  sizeof(double),
  sizeof(long double),
};

fcctype_t fccode_type(char *f)
{
  char *p;

  p = strchr(f, '%');
  while (p && *(p+1)=='%')
    if (*(p+2)=='\0')
      return FCCTYPE_VOID;
    else
      p = strchr(p+2, '%');
  if (!p)
    return FCCTYPE_VOID;
  else if (strstr(p,"Le")||strstr(p,"Lf")||strstr(p,"Lg"))
    return FCCTYPE_QUADRUPLE;
  else if (strstr(p,"le")||strstr(p,"lf")||strstr(p,"lg"))
    return FCCTYPE_DOUBLE;
  else if (strstr(p,"e")||strstr(p,"f")||strstr(p,"g"))
    return FCCTYPE_FLOAT;
  else if (strstr(p,"lo")||strstr(p,"lu")||strstr(p,"lx"))
    return FCCTYPE_ULONG;
  else if (strstr(p,"ho")||strstr(p,"hu")||strstr(p,"hx"))
    return FCCTYPE_USHORT;
  else if (strstr(p,"o")||strstr(p,"u")||strstr(p,"x"))
    return FCCTYPE_UINT;
  else if (strstr(p,"ld")||strstr(p,"li"))
    return FCCTYPE_LONG;
  else if (strstr(p,"hd")||strstr(p,"hi"))
    return FCCTYPE_SHORT;
  else if (strstr(p,"d")||strstr(p,"i"))
    return FCCTYPE_INT;
  else if (strstr(p,"ws"))
    return FCCTYPE_WSTRING;
  else if (strstr(p,"s"))
    return FCCTYPE_STRING;
  else if (strstr(p,"wc"))
    return FCCTYPE_WCHAR;
  else if (strstr(p,"c"))
    return FCCTYPE_CHAR;
  else
    return FCCTYPE_VOID;
}

size_t fccode_size(char *f)
{
  return fccode_sizes[fccode_type(f)];
}

struct one_dimension_printf_style {
  char *begin, *delimiter, *end;
};
struct two_dimension_printf_style {
  struct one_dimension_printf_style row, column;
};
struct three_dimension_printf_style {
  struct one_dimension_printf_style row, column, depth;
};

static struct one_dimension_printf_style vector_fprintf_style = {
  "", " ", ""
};
static struct one_dimension_printf_style vector_sprintf_style = {
  "", ",", ""
};

static struct two_dimension_printf_style matrix_fprintf_style = {
  { "", "\n", "" },
  { "", " ", "" },
};
static struct two_dimension_printf_style matrix_sprintf_style = {
  { "", ",", "" },
  { "(", ",", ")" },
};

static struct three_dimension_printf_style f3tensor_fprintf_style = {
  { "", "\n\n", "" },
  { "", "\n", "" },
  { "", " ", "" },
};
static struct three_dimension_printf_style f3tensor_sprintf_style = {
  { "", ",", "" },
  { "{", ",", "}" },
  { "(", ",", ")" },
};

struct one_dimension_scanf_style {
  char *begin, *delimiter, *end;
};
struct two_dimension_scanf_style {
  char *begin, *row_delimiter, *column_delimiter, *end;
};
struct three_dimension_scanf_style {
  char *begin, *row_delimiter, *column_delimiter, *depth_delimiter, *end;
};

#if 0 /* UNCOMPLETED */
static struct one_dimension_scanf_style vector_fscanf_style = {
  "", " ", ""
};
#endif
static struct one_dimension_scanf_style vector_sscanf_style = {
  "", ",", ""
};

#if 0 /* UNCOMPLETED */
static struct two_dimension_scanf_style matrix_fscanf_style = {
  "", "\n", " ", ""
};
#endif
static struct two_dimension_scanf_style matrix_sscanf_style = {
  "(", "),(", ",", ")"
};

#if 0 /* UNCOMPLETED */
static struct three_dimension_scanf_style f3tensor_fscanf_style = {
  "", "\n\n", "\n", " ", ""
};
#endif
static struct three_dimension_scanf_style f3tensor_sscanf_style = {
  "{(", ")},{(", "),(", ",", ")}"
};

int (vector_fprintf)(size_t size, FILE *fp, char *f, void *v, long nl, long nh)
{
  long i;
  int c = 0;
  fcctype_t type = fccode_type(f);
  static struct one_dimension_printf_style *style = &vector_fprintf_style;

  if (type == FCCTYPE_VOID)
    return c;
  c += fprintf(fp, "%s", style->begin);
  for (i=nl; i<=nh; i++) {
    switch (type) {
    case FCCTYPE_CHAR:
      c += fprintf(fp, f, *(char *)(v+size*i));
      break;
    case FCCTYPE_WCHAR:
      c += fprintf(fp, f, *(wchar_t *)(v+size*i));
      break;
    case FCCTYPE_STRING:
      c += fprintf(fp, f, *(char **)(v+size*i));
      break;
    case FCCTYPE_WSTRING:
      c += fprintf(fp, f, *(wchar_t **)(v+size*i));
      break;
    case FCCTYPE_INT:
      c += fprintf(fp, f, *(int *)(v+size*i));
      break;
    case FCCTYPE_SHORT:
      c += fprintf(fp, f, *(short int *)(v+size*i));
      break;
    case FCCTYPE_LONG:
      c += fprintf(fp, f, *(long int *)(v+size*i));
      break;
    case FCCTYPE_UINT:
      c += fprintf(fp, f, *(unsigned int *)(v+size*i));
      break;
    case FCCTYPE_USHORT:
      c += fprintf(fp, f, *(unsigned short int *)(v+size*i));
      break;
    case FCCTYPE_ULONG:
      c += fprintf(fp, f, *(unsigned long int *)(v+size*i));
      break;
    case FCCTYPE_FLOAT:
    case FCCTYPE_DOUBLE:
      if (size == fccode_sizes[FCCTYPE_FLOAT])
        c += fprintf(fp, f, *(float *)(v+size*i));
      else
        c += fprintf(fp, f, *(double *)(v+size*i));
      break;
    case FCCTYPE_QUADRUPLE:
      c += fprintf(fp, f, *(long double *)(v+size*i));
      break;
    default:
      break;
    }
    if (i!=nh) {
      c += fprintf(fp, "%s", style->delimiter);
    }
  }
  c += fprintf(fp, "%s", style->end);
  return c;
}

int (vector_printf)(size_t size, char *f, void *v, long nl, long nh)
{
  return (vector_fprintf)(size, stdout, f, v, nl, nh);
}

int (vector_sprintf)(size_t size, char *buf, char *f, void *v, long nl, long nh)
{
  long i;
  int c = 0;
  fcctype_t type = fccode_type(f);
  char *p = buf;
  static struct one_dimension_printf_style *style = &vector_sprintf_style;

  if (type == FCCTYPE_VOID)
    return c;
  c += sprintf(p, "%s", style->begin);
  p = strchr(p, '\0');
  for (i=nl; i<=nh; i++) {
    switch (type) {
    case FCCTYPE_CHAR:
      c += sprintf(p, f, *(char *)(v+size*i));
      break;
    case FCCTYPE_WCHAR:
      c += sprintf(p, f, *(wchar_t *)(v+size*i));
      break;
    case FCCTYPE_STRING:
      c += sprintf(p, f, *(char **)(v+size*i));
      break;
    case FCCTYPE_WSTRING:
      c += sprintf(p, f, *(wchar_t **)(v+size*i));
      break;
    case FCCTYPE_INT:
      c += sprintf(p, f, *(int *)(v+size*i));
      break;
    case FCCTYPE_SHORT:
      c += sprintf(p, f, *(short int *)(v+size*i));
      break;
    case FCCTYPE_LONG:
      c += sprintf(p, f, *(long int *)(v+size*i));
      break;
    case FCCTYPE_UINT:
      c += sprintf(p, f, *(unsigned int *)(v+size*i));
      break;
    case FCCTYPE_USHORT:
      c += sprintf(p, f, *(unsigned short int *)(v+size*i));
      break;
    case FCCTYPE_ULONG:
      c += sprintf(p, f, *(unsigned long int *)(v+size*i));
      break;
    case FCCTYPE_FLOAT:
    case FCCTYPE_DOUBLE:
      if (size == fccode_sizes[FCCTYPE_FLOAT])
        c += sprintf(p, f, *(float *)(v+size*i));
      else
        c += sprintf(p, f, *(double *)(v+size*i));
      break;
    case FCCTYPE_QUADRUPLE:
      c += sprintf(p, f, *(long double *)(v+size*i));
      break;
    default:
      break;
    }
    p = strchr(p, '\0');
    if (i!=nh) {
      c += sprintf(p, "%s", style->delimiter);
      p = strchr(p, '\0');
    }
  }
  c += sprintf(p, "%s", style->end);
  p = strchr(p, '\0');
  return c;
}

int (vector_sscanf)(size_t size, char *buf, char *f, void **v, long nl, long *nh)
{
  long i = nl, l;
  int c = 0;
  fcctype_t type = fccode_type(f);
  char *p = buf;
  static struct one_dimension_scanf_style *style = &vector_sscanf_style;

  if (type == FCCTYPE_VOID)
    return c;
  while (p && *p) {
    if (style->begin && *style->begin
        && strncmp(p,style->begin,l=strlen(style->begin))==0) {
      p = p + l;
    } else if (style->delimiter && *style->delimiter
               && strncmp(p,style->delimiter,l=strlen(style->delimiter))==0) {
      p = p + l;
    } else if (style->end && *style->end
               && strncmp(p,style->end,l=strlen(style->end))==0) {
      p = p + l;
      break;
    } else if (p == buf && (!style->begin||(style->begin&&!*style->begin))) {
      ;
    } else {
      p++;
      continue;
    }
    if (*v && i<=*nh)
      switch (type) {
      case FCCTYPE_CHAR:
        c += sscanf(p, f, (char *)(*v+size*i));
        break;
      case FCCTYPE_WCHAR:
        c += sscanf(p, f, (wchar_t *)(*v+size*i));
        break;
      case FCCTYPE_STRING:
        c += sscanf(p, f, (char **)(*v+size*i));
        break;
      case FCCTYPE_WSTRING:
        c += sscanf(p, f, (wchar_t **)(*v+size*i));
        break;
      case FCCTYPE_INT:
        c += sscanf(p, f, (int *)(*v+size*i));
        break;
      case FCCTYPE_SHORT:
        c += sscanf(p, f, (short int *)(*v+size*i));
        break;
      case FCCTYPE_LONG:
        c += sscanf(p, f, (long int *)(*v+size*i));
        break;
      case FCCTYPE_UINT:
        c += sscanf(p, f, (unsigned int *)(*v+size*i));
        break;
      case FCCTYPE_USHORT:
        c += sscanf(p, f, (unsigned short int *)(*v+size*i));
        break;
      case FCCTYPE_ULONG:
        c += sscanf(p, f, (unsigned long int *)(*v+size*i));
        break;
      case FCCTYPE_FLOAT:
        c += sscanf(p, f, (float *)(*v+size*i));
        break;
      case FCCTYPE_DOUBLE:
        c += sscanf(p, f, (double *)(*v+size*i));
        break;
      case FCCTYPE_QUADRUPLE:
        c += sscanf(p, f, (long double *)(*v+size*i));
        break;
      default:
        break;
      }
    i++;
    if (p == buf) p++;
  }
  if (*v == NULL) {
    *nh = i-1;
    *v = (vector_alloc)(size, nl, *nh);
    if (v)
      return (vector_sscanf)(size, buf, f, v, nl, nh);
  }
  return c;
}

#if 0 /* UNCOMPLETED */
int (vector_fscanf)(size_t size, FILE *fp, char *f, void **v, long nl, long *nh)
{
}

int (vector_scanf)(size_t size, char *f, void **v, long nl, long *nh)
{
}
#endif

int (matrix_fprintf)(size_t size, FILE *fp, char *f, void **m, long nrl, long nrh, long ncl, long nch)
{
  long i, j;
  int c = 0;
  fcctype_t type = fccode_type(f);
  static struct two_dimension_printf_style *style = &matrix_fprintf_style;

  if (type == FCCTYPE_VOID)
    return c;
  c += fprintf(fp, "%s", style->row.begin);
  for (i=nrl; i<=nrh; i++) {
    c += fprintf(fp, "%s", style->column.begin);
    for (j=ncl; j<=nch; j++) {
      switch (type) {
      case FCCTYPE_CHAR:
        c += fprintf(fp, f, *(char *)(*(m+i)+size*j));
        break;
      case FCCTYPE_WCHAR:
        c += fprintf(fp, f, *(wchar_t *)(*(m+i)+size*j));
        break;
      case FCCTYPE_STRING:
        c += fprintf(fp, f, *(char **)(*(m+i)+size*j));
        break;
      case FCCTYPE_WSTRING:
        c += fprintf(fp, f, *(wchar_t **)(*(m+i)+size*j));
        break;
      case FCCTYPE_INT:
        c += fprintf(fp, f, *(int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_SHORT:
        c += fprintf(fp, f, *(short int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_LONG:
        c += fprintf(fp, f, *(long int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_UINT:
        c += fprintf(fp, f, *(unsigned int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_USHORT:
        c += fprintf(fp, f, *(unsigned short int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_ULONG:
        c += fprintf(fp, f, *(unsigned long int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_FLOAT:
      case FCCTYPE_DOUBLE:
        if (size == fccode_sizes[FCCTYPE_FLOAT])
          c += fprintf(fp, f, *(float *)(*(m+i)+size*j));
        else
          c += fprintf(fp, f, *(double *)(*(m+i)+size*j));
        break;
      case FCCTYPE_QUADRUPLE:
        c += fprintf(fp, f, *(long double *)(*(m+i)+size*j));
        break;
      default:
        break;
      }
      if (j!=nch) {
        c += fprintf(fp, "%s", style->column.delimiter);
      }
    }
    c += fprintf(fp, "%s", style->column.end);
    if (i!=nrh) {
      c += fprintf(fp, "%s", style->row.delimiter);
    }
  }
  c += fprintf(fp, "%s", style->row.end);
  return c;
}

int (matrix_printf)(size_t size, char *f, void **m, long nrl, long nrh, long ncl, long nch)
{
  return (matrix_fprintf)(size, stdout, f, m, nrl, nrh, ncl, nch);
}

int (matrix_sprintf)(size_t size, char *buf, char *f, void **m, long nrl, long nrh, long ncl, long nch)
{
  long i, j;
  int c = 0;
  fcctype_t type = fccode_type(f);
  char *p = buf;
  static struct two_dimension_printf_style *style = &matrix_sprintf_style;

  if (type == FCCTYPE_VOID)
    return c;
  c += sprintf(p, "%s", style->row.begin);
  p = strchr(p, '\0');
  for (i=nrl; i<=nrh; i++) {
    c += sprintf(p, "%s", style->column.begin);
    p = strchr(p, '\0');
    for (j=ncl; j<=nch; j++) {
      switch (type) {
      case FCCTYPE_CHAR:
        c += sprintf(p, f, *(char *)(*(m+i)+size*j));
        break;
      case FCCTYPE_WCHAR:
        c += sprintf(p, f, *(wchar_t *)(*(m+i)+size*j));
        break;
      case FCCTYPE_STRING:
        c += sprintf(p, f, *(char **)(*(m+i)+size*j));
        break;
      case FCCTYPE_WSTRING:
        c += sprintf(p, f, *(wchar_t **)(*(m+i)+size*j));
        break;
      case FCCTYPE_INT:
        c += sprintf(p, f, *(int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_SHORT:
        c += sprintf(p, f, *(short int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_LONG:
        c += sprintf(p, f, *(long int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_UINT:
        c += sprintf(p, f, *(unsigned int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_USHORT:
        c += sprintf(p, f, *(unsigned short int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_ULONG:
        c += sprintf(p, f, *(unsigned long int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_FLOAT:
      case FCCTYPE_DOUBLE:
        if (size == fccode_sizes[FCCTYPE_FLOAT])
          c += sprintf(p, f, *(float *)(*(m+i)+size*j));
        else
          c += sprintf(p, f, *(double *)(*(m+i)+size*j));
        break;
      case FCCTYPE_QUADRUPLE:
        c += sprintf(p, f, *(long double *)(*(m+i)+size*j));
        break;
      default:
        break;
      }
      p = strchr(p, '\0');
      if (j!=nch) {
        c += sprintf(p, "%s", style->column.delimiter);
        p = strchr(p, '\0');
      }
    }
    c += sprintf(p, "%s", style->column.end);
    p = strchr(p, '\0');
    if (i!=nrh) {
      c += sprintf(p, "%s", style->row.delimiter);
      p = strchr(p, '\0');
    }
  }
  c += sprintf(p, "%s", style->row.end);
  p = strchr(p, '\0');
  return c;
}

int (matrix_sscanf)(size_t size, char *buf, char *f, void ***m, long nrl, long *nrh, long ncl, long *nch)
{
  long i = nrl, j = ncl, l;
  int c = 0;
  fcctype_t type = fccode_type(f);
  char *p = buf;
  static struct two_dimension_scanf_style *style = &matrix_sscanf_style;

  if (type == FCCTYPE_VOID)
    return c;
  while (p && *p) {
    if (style->begin && *style->begin
        && strncmp(p,style->begin,l=strlen(style->begin))==0) {
      p = p + l;
    } else if (style->row_delimiter && *style->row_delimiter
               && strncmp(p,style->row_delimiter,l=strlen(style->row_delimiter))==0) {
      p = p + l;
      j = ncl;
      i++;
      if (*m && i > *nrh)
        break;
    } else if (style->column_delimiter && *style->column_delimiter
               && strncmp(p,style->column_delimiter,l=strlen(style->column_delimiter))==0) {
      p = p + l;
    } else if (style->end && *style->end
               && strncmp(p,style->end,l=strlen(style->end))==0) {
      p = p + l;
      break;
    } else if (p == buf && (!style->begin||(style->begin&&!*style->begin))) {
      ;
    } else {
      p++;
      continue;
    }
    if (*m && i<=*nrh && j<=*nch)
      switch (type) {
      case FCCTYPE_CHAR:
        c += sscanf(p, f, (char *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_WCHAR:
        c += sscanf(p, f, (wchar_t *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_STRING:
        c += sscanf(p, f, (char **)(*(*m+i)+size*j));
        break;
      case FCCTYPE_WSTRING:
        c += sscanf(p, f, (wchar_t **)(*(*m+i)+size*j));
        break;
      case FCCTYPE_INT:
        c += sscanf(p, f, (int *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_SHORT:
        c += sscanf(p, f, (short int *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_LONG:
        c += sscanf(p, f, (long int *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_UINT:
        c += sscanf(p, f, (unsigned int *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_USHORT:
        c += sscanf(p, f, (unsigned short int *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_ULONG:
        c += sscanf(p, f, (unsigned long int *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_FLOAT:
        c += sscanf(p, f, (float *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_DOUBLE:
        c += sscanf(p, f, (double *)(*(*m+i)+size*j));
        break;
      case FCCTYPE_QUADRUPLE:
        c += sscanf(p, f, (long double *)(*(*m+i)+size*j));
        break;
      default:
        break;
      }
    j++;
    if (p == buf) p++;
  }
  if (*m == NULL) {
    *nrh = i;
    *nch = j-1;
    *m = (matrix_alloc)(size, nrl, *nrh, ncl, *nch);
    if (m)
      return (matrix_sscanf)(size, buf, f, m, nrl, nrh, ncl, nch);
  }
  return c;
}

#if 0 /* UNCOMPLETED */
int (matrix_fscanf)(size_t size, FILE *fp, char *f, void ***m, long nrl, long *nrh, long ncl, long *nch)
{
}

int (matrix_scanf)(size_t size, char *f, void ***m, long nrl, long *nrh, long ncl, long *nch)
{
}
#endif

int (matrixT_fprintf)(size_t size, FILE *fp, char *f, void **m, long nrl, long nrh, long ncl, long nch)
{
  long i, j;
  int c = 0;
  fcctype_t type = fccode_type(f);
  static struct two_dimension_printf_style *style = &matrix_fprintf_style;

  if (type == FCCTYPE_VOID)
    return c;
  c += fprintf(fp, "%s", style->row.begin);
  for (j=ncl; j<=nch; j++) {
    c += fprintf(fp, "%s", style->column.begin);
    for (i=nrl; i<=nrh; i++) {
      switch (type) {
      case FCCTYPE_CHAR:
        c += fprintf(fp, f, *(char *)(*(m+i)+size*j));
        break;
      case FCCTYPE_WCHAR:
        c += fprintf(fp, f, *(wchar_t *)(*(m+i)+size*j));
        break;
      case FCCTYPE_STRING:
        c += fprintf(fp, f, *(char **)(*(m+i)+size*j));
        break;
      case FCCTYPE_WSTRING:
        c += fprintf(fp, f, *(wchar_t **)(*(m+i)+size*j));
        break;
      case FCCTYPE_INT:
        c += fprintf(fp, f, *(int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_SHORT:
        c += fprintf(fp, f, *(short int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_LONG:
        c += fprintf(fp, f, *(long int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_UINT:
        c += fprintf(fp, f, *(unsigned int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_USHORT:
        c += fprintf(fp, f, *(unsigned short int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_ULONG:
        c += fprintf(fp, f, *(unsigned long int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_FLOAT:
      case FCCTYPE_DOUBLE:
        if (size == fccode_sizes[FCCTYPE_FLOAT])
          c += fprintf(fp, f, *(float *)(*(m+i)+size*j));
        else
          c += fprintf(fp, f, *(double *)(*(m+i)+size*j));
        break;
      case FCCTYPE_QUADRUPLE:
        c += fprintf(fp, f, *(long double *)(*(m+i)+size*j));
        break;
      default:
        break;
      }
      if (i!=nrh) {
        c += fprintf(fp, "%s", style->column.delimiter);
      }
    }
    c += fprintf(fp, "%s", style->column.end);
    if (j!=nch) {
      c += fprintf(fp, "%s", style->row.delimiter);
    }
  }
  c += fprintf(fp, "%s", style->row.end);
  return c;
}

int (matrixT_printf)(size_t size, char *f, void **m, long nrl, long nrh, long ncl, long nch)
{
  return (matrixT_fprintf)(size, stdout, f, m, nrl, nrh, ncl, nch);
}

int (matrixT_sprintf)(size_t size, char *buf, char *f, void **m, long nrl, long nrh, long ncl, long nch)
{
  long i, j;
  int c = 0;
  fcctype_t type = fccode_type(f);
  char *p = buf;
  static struct two_dimension_printf_style *style = &matrix_sprintf_style;

  if (type == FCCTYPE_VOID)
    return c;
  c += sprintf(p, "%s", style->row.begin);
  p = strchr(p, '\0');
  for (j=ncl; j<=nch; j++) {
    c += sprintf(p, "%s", style->column.begin);
    p = strchr(p, '\0');
    for (i=nrl; i<=nrh; i++) {
      switch (type) {
      case FCCTYPE_CHAR:
        c += sprintf(p, f, *(char *)(*(m+i)+size*j));
        break;
      case FCCTYPE_WCHAR:
        c += sprintf(p, f, *(wchar_t *)(*(m+i)+size*j));
        break;
      case FCCTYPE_STRING:
        c += sprintf(p, f, *(char **)(*(m+i)+size*j));
        break;
      case FCCTYPE_WSTRING:
        c += sprintf(p, f, *(wchar_t **)(*(m+i)+size*j));
        break;
      case FCCTYPE_INT:
        c += sprintf(p, f, *(int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_SHORT:
        c += sprintf(p, f, *(short int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_LONG:
        c += sprintf(p, f, *(long int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_UINT:
        c += sprintf(p, f, *(unsigned int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_USHORT:
        c += sprintf(p, f, *(unsigned short int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_ULONG:
        c += sprintf(p, f, *(unsigned long int *)(*(m+i)+size*j));
        break;
      case FCCTYPE_FLOAT:
      case FCCTYPE_DOUBLE:
        if (size == fccode_sizes[FCCTYPE_FLOAT])
          c += sprintf(p, f, *(float *)(*(m+i)+size*j));
        else
          c += sprintf(p, f, *(double *)(*(m+i)+size*j));
        break;
      case FCCTYPE_QUADRUPLE:
        c += sprintf(p, f, *(long double *)(*(m+i)+size*j));
        break;
      default:
        break;
      }
      p = strchr(p, '\0');
      if (i!=nrh) {
        c += sprintf(p, "%s", style->column.delimiter);
        p = strchr(p, '\0');
      }
    }
    c += sprintf(p, "%s", style->column.end);
    p = strchr(p, '\0');
    if (j!=nch) {
      c += sprintf(p, "%s", style->row.delimiter);
      p = strchr(p, '\0');
    }
  }
  c += sprintf(p, "%s", style->row.end);
  p = strchr(p, '\0');
  return c;
}

int (matrixT_sscanf)(size_t size, char *buf, char *f, void ***m, long ncl, long *nch, long nrl, long *nrh/*long nrl, long *nrh, long ncl, long *nch*/)
{
  long i = nrl, j = ncl, l;
  int c = 0;
  fcctype_t type = fccode_type(f);
  char *p = buf;
  static struct two_dimension_scanf_style *style = &matrix_sscanf_style;

  if (type == FCCTYPE_VOID)
    return c;
  while (p && *p) {
    if (style->begin && *style->begin
        && strncmp(p,style->begin,l=strlen(style->begin))==0) {
      p = p + l;
    } else if (style->row_delimiter && *style->row_delimiter
               && strncmp(p,style->row_delimiter,l=strlen(style->row_delimiter))==0) {
      p = p + l;
      j = ncl;
      i++;
      if (*m && i > *nrh)
        break;
    } else if (style->column_delimiter && *style->column_delimiter
               && strncmp(p,style->column_delimiter,l=strlen(style->column_delimiter))==0) {
      p = p + l;
    } else if (style->end && *style->end
               && strncmp(p,style->end,l=strlen(style->end))==0) {
      p = p + l;
      break;
    } else if (p == buf && (!style->begin||(style->begin&&!*style->begin))) {
      ;
    } else {
      p++;
      continue;
    }
    if (*m && i<=*nrh && j<=*nch)
      switch (type) {
      case FCCTYPE_CHAR:
        c += sscanf(p, f, (char *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_WCHAR:
        c += sscanf(p, f, (wchar_t *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_STRING:
        c += sscanf(p, f, (char **)(*(*m+j)+size*i));
        break;
      case FCCTYPE_WSTRING:
        c += sscanf(p, f, (wchar_t **)(*(*m+j)+size*i));
        break;
      case FCCTYPE_INT:
        c += sscanf(p, f, (int *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_SHORT:
        c += sscanf(p, f, (short int *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_LONG:
        c += sscanf(p, f, (long int *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_UINT:
        c += sscanf(p, f, (unsigned int *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_USHORT:
        c += sscanf(p, f, (unsigned short int *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_ULONG:
        c += sscanf(p, f, (unsigned long int *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_FLOAT:
        c += sscanf(p, f, (float *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_DOUBLE:
        c += sscanf(p, f, (double *)(*(*m+j)+size*i));
        break;
      case FCCTYPE_QUADRUPLE:
        c += sscanf(p, f, (long double *)(*(*m+j)+size*i));
        break;
      default:
        break;
      }
    j++;
    if (p == buf) p++;
  }
  if (*m == NULL) {
    *nrh = i;
    *nch = j-1;
    *m = (matrix_alloc)(size, nrl, *nrh, ncl, *nch);
    if (m)
      /*return (matrix_sscanf)(size, buf, f, m, nrl, nrh, ncl, nch);*/
      return (matrixT_sscanf)(size, buf, f, m, ncl, nch, nrl, nrh);
  }
  return c;
}

#if 0 /* UNCOMPLETED */
int (matrixT_fscanf)(size_t size, FILE *fp, char *f, void ***m, long ncl, long *nch, long nrl, long *nrh/*long nrl, long *nrh, long ncl, long *nch*/)
{
}

int (matrixT_scanf)(size_t size, char *f, void ***m, long ncl, long *nch, long nrl, long *nrh/*long nrl, long *nrh, long ncl, long *nch*/)
{
}
#endif

int (f3tensor_fprintf)(size_t size, FILE *fp, char *f, void ***t, long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
{
  long i, j, k;
  int c = 0;
  fcctype_t type = fccode_type(f);
  static struct three_dimension_printf_style *style = &f3tensor_fprintf_style;

  if (type == FCCTYPE_VOID)
    return c;
  c += fprintf(fp, "%s", style->row.begin);
  for (i=nrl; i<=nrh; i++) {
    c += fprintf(fp, "%s", style->column.begin);
    for (j=ncl; j<=nch; j++) {
      c += fprintf(fp, "%s", style->depth.begin);
      for (k=ndl; k<=ndh; k++) {
        switch (type) {
        case FCCTYPE_CHAR:
          c += fprintf(fp, f, *(char *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_WCHAR:
          c += fprintf(fp, f, *(wchar_t *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_STRING:
          c += fprintf(fp, f, *(char **)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_WSTRING:
          c += fprintf(fp, f, *(wchar_t **)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_INT:
          c += fprintf(fp, f, *(int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_SHORT:
          c += fprintf(fp, f, *(short int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_LONG:
          c += fprintf(fp, f, *(long int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_UINT:
          c += fprintf(fp, f, *(unsigned int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_USHORT:
          c += fprintf(fp, f, *(unsigned short int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_ULONG:
          c += fprintf(fp, f, *(unsigned long int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_FLOAT:
        case FCCTYPE_DOUBLE:
          if (size == fccode_sizes[FCCTYPE_FLOAT])
            c += fprintf(fp, f, *(float *)(*(*(t+i)+j)+size*k));
          else
            c += fprintf(fp, f, *(double *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_QUADRUPLE:
          c += fprintf(fp, f, *(long double *)(*(*(t+i)+j)+size*k));
          break;
        default:
          break;
        }
        if (k!=ndh) {
          c += fprintf(fp, "%s", style->depth.delimiter);
        }
      }
      c += fprintf(fp, "%s", style->depth.end);
      if (j!=nch) {
        c += fprintf(fp, "%s", style->column.delimiter);
      }
    }
    c += fprintf(fp, "%s", style->column.end);
    if (i!=nrh) {
      c += fprintf(fp, "%s", style->row.delimiter);
    }
  }
  c += fprintf(fp, "%s", style->row.end);
  return c;
}

int (f3tensor_printf)(size_t size, char *f, void ***t, long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
{
  return (f3tensor_fprintf)(size, stdout, f, t, nrl, nrh, ncl, nch, ndl, ndh);
}

int (f3tensor_sprintf)(size_t size, char *buf, char *f, void ***t, long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
{
  long i, j, k;
  int c = 0;
  fcctype_t type = fccode_type(f);
  char *p = buf;
  static struct three_dimension_printf_style *style = &f3tensor_sprintf_style;

  if (type == FCCTYPE_VOID)
    return c;
  c += sprintf(p, "%s", style->row.begin);
  p = strchr(p, '\0');
  for (i=nrl; i<=nrh; i++) {
    c += sprintf(p, "%s", style->column.begin);
    p = strchr(p, '\0');
    for (j=ncl; j<=nch; j++) {
      c += sprintf(p, "%s", style->depth.begin);
      p = strchr(p, '\0');
      for (k=ndl; k<=ndh; k++) {
        switch (type) {
        case FCCTYPE_CHAR:
          c += sprintf(p, f, *(char *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_WCHAR:
          c += sprintf(p, f, *(wchar_t *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_STRING:
          c += sprintf(p, f, *(char **)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_WSTRING:
          c += sprintf(p, f, *(wchar_t **)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_INT:
          c += sprintf(p, f, *(int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_SHORT:
          c += sprintf(p, f, *(short int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_LONG:
          c += sprintf(p, f, *(long int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_UINT:
          c += sprintf(p, f, *(unsigned int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_USHORT:
          c += sprintf(p, f, *(unsigned short int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_ULONG:
          c += sprintf(p, f, *(unsigned long int *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_FLOAT:
        case FCCTYPE_DOUBLE:
          if (size == fccode_sizes[FCCTYPE_FLOAT])
            c += sprintf(p, f, *(float *)(*(*(t+i)+j)+size*k));
          else
            c += sprintf(p, f, *(double *)(*(*(t+i)+j)+size*k));
          break;
        case FCCTYPE_QUADRUPLE:
          c += sprintf(p, f, *(long double *)(*(*(t+i)+j)+size*k));
          break;
        default:
          break;
        }
        p = strchr(p, '\0');
        if (k!=ndh) {
          c += sprintf(p, "%s", style->depth.delimiter);
          p = strchr(p, '\0');
        }
      }
      c += sprintf(p, "%s", style->depth.end);
      p = strchr(p, '\0');
      if (j!=nch) {
        c += sprintf(p, "%s", style->column.delimiter);
        p = strchr(p, '\0');
      }
    }
    c += sprintf(p, "%s", style->column.end);
    p = strchr(p, '\0');
    if (i!=nrh) {
      c += sprintf(p, "%s", style->row.delimiter);
      p = strchr(p, '\0');
    }
  }
  c += sprintf(p, "%s", style->row.end);
  p = strchr(p, '\0');
  return c;
}

int (f3tensor_sscanf)(size_t size, char *buf, char *f, void ****t, long nrl, long *nrh, long ncl, long *nch, long ndl, long *ndh)
{
  long i = nrl, j = ncl, k = ndl, l;
  int c = 0;
  fcctype_t type = fccode_type(f);
  char *p = buf;
  static struct three_dimension_scanf_style *style = &f3tensor_sscanf_style;

  if (type == FCCTYPE_VOID)
    return c;
  while (p && *p) {
    if (style->begin && *style->begin
        && strncmp(p,style->begin,l=strlen(style->begin))==0) {
      p = p + l;
    } else if (style->row_delimiter && *style->row_delimiter
               && strncmp(p,style->row_delimiter,l=strlen(style->row_delimiter))==0) {
      p = p + l;
      k = ndl;
      j = ncl;
      i++;
      if (*t && i > *nrh)
        break;
    } else if (style->column_delimiter && *style->column_delimiter
               && strncmp(p,style->column_delimiter,l=strlen(style->column_delimiter))==0) {
      p = p + l;
      k = ndl;
      j++;
      if (*t && j > *nch) {
	j = ncl;
	i++;
	if (*t && i > *nrh)
	  break;
      }
    } else if (style->depth_delimiter && *style->depth_delimiter
               && strncmp(p,style->depth_delimiter,l=strlen(style->depth_delimiter))==0) {
      p = p + l;
    } else if (style->end && *style->end
               && strncmp(p,style->end,l=strlen(style->end))==0) {
      p = p + l;
      break;
    } else if (p == buf && (!style->begin||(style->begin&&!*style->begin))) {
      ;
    } else {
      p++;
      continue;
    }
    if (*t && i<=*nrh && j<=*nch && k<=*ndh)
      switch (type) {
      case FCCTYPE_CHAR:
        c += sscanf(p, f, (char *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_WCHAR:
        c += sscanf(p, f, (wchar_t *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_STRING:
        c += sscanf(p, f, (char **)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_WSTRING:
        c += sscanf(p, f, (wchar_t **)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_INT:
        c += sscanf(p, f, (int *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_SHORT:
        c += sscanf(p, f, (short int *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_LONG:
        c += sscanf(p, f, (long int *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_UINT:
        c += sscanf(p, f, (unsigned int *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_USHORT:
        c += sscanf(p, f, (unsigned short int *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_ULONG:
        c += sscanf(p, f, (unsigned long int *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_FLOAT:
        c += sscanf(p, f, (float *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_DOUBLE:
        c += sscanf(p, f, (double *)(*(*(*t+i)+j)+size*k));
        break;
      case FCCTYPE_QUADRUPLE:
        c += sscanf(p, f, (long double *)(*(*(*t+i)+j)+size*k));
        break;
      default:
        break;
      }
    k++;
    if (p == buf) p++;
  }
  if (*t == NULL) {
    *nrh = i;
    *nch = j;
    *ndh = k-1;
    *t = (f3tensor_alloc)(size, nrl, *nrh, ncl, *nch, ndl, *ndh);
    if (t)
      return (f3tensor_sscanf)(size, buf, f, t, nrl, nrh, ncl, nch, ndl, ndh);
  }
  return c;
}

#if 0 /* UNCOMPLETED */
int (f3tensor_fscanf)(size_t size, FILE *fp, char *f, void ****t, long nrl, long *nrh, long ncl, long *nch, long ndl, long *ndh)
{
}

int (f3tensor_scanf)(size_t size, char *f, void ****t, long nrl, long *nrh, long ncl, long *nch, long ndl, long *ndh)
{
}
#endif
